; boot.asm
[Bits 16]
org	0

jmp	entry_point					       ; jump to bootloader entry point

;##############################################################################
; "FAT 12" file system is build up by four areas at Floppy Disk:
; - Boot Sector (also called BIOS Parameter Block, BPB)
; - File Allocation Table, FAT 
; - Root Directory
; - File and Directory Region

; boot sector info block
; begins 3 bytes from start. A far jump needs 3 bytes. 

OperatingSystemName db "PrettyOS"      ;  8 byte
BytesPerSec         dw 512
SecPerClus          db 1
ReservedSec         dw 1
NumFATs             db 2
RootEntries         dw 224
TotSec              dw 2880
MediaType           db 0xF0
FATSize             dw 9
SecPerTrack         dw 18
NumHeads            dw 2
HiddenSec           dd 0
TotSec32            dd 0
DriveNum            db 0
Reserved            db 0
BootSig             db 0x29
VolSerialNum        dd 0xD00FC0DE
VolLabel            db "PrettyOS   "   ; 11 byte
FileSys             db "FAT12   "      ;  8 byte
;##############################################################################


;******************************************************************************
;	bootloader entry point
;******************************************************************************
entry_point:
    mov     ax, 0x07C0		; setup registers to point to our segment
    mov     ds, ax
    mov     es, ax
    mov     fs, ax
    mov     gs, ax
    xor     ax, ax			; set the stack
    mov     ss, ax
    mov     sp, 0xFFFF
          
    mov  [bootdevice], dl

;******************************************************************************
;   Load root directory table
;******************************************************************************
LOAD_ROOT:
    ; compute size of root directory and store in "cx"
    xor cx, cx
    xor dx, dx
    mov ax, 0x0020                           ; 32 byte directory entry
    mul WORD [RootEntries]                   ; total size of directory
    div WORD [BytesPerSec]                   ; sectors used by directory
    xchg ax, cx
          
    ; compute location of root directory and store in "ax"
    mov al, BYTE [NumFATs]                    ; number of FATs
    mul WORD [FATSize]                        ; sectors used by FATs
    add ax, WORD [ReservedSec]                ; adjust for bootsector
    mov WORD [datasector], ax                 ; base of root directory
    add WORD [datasector], cx
          
    ; read root directory into memory (7C00:0200)
    mov bx, 0x0200                            ; copy root dir above bootcode
    call ReadSectors

;******************************************************************************
;   Find stage2 bootloader
;******************************************************************************
    ; browse root directory for binary image
    mov cx, WORD [RootEntries]                ; load loop counter
    mov di, 0x0200                            ; locate first root entry
.LOOP:
    push cx
    mov cx, 0x000B                            ; name has 11 characters 
    mov si, ImageName                         ; look for this image name
    push di
    rep cmpsb                                 ; test for entry match
    pop di
    je LOAD_FAT
    pop cx
    add di, 0x0020                            ; queue next directory entry
    loop .LOOP
    jmp FAILURE

;******************************************************************************
;   Load File Allocation Table (FAT)
;******************************************************************************

LOAD_FAT:
    ; save starting cluster of boot image
    mov dx, WORD [di + 0x001A]
    mov WORD [cluster], dx                  ; file's first cluster
          
    ; compute size of FAT and store in "cx"
    xor ax, ax
    mov al, BYTE [NumFATs]                  ; number of FATs
    mul WORD [FATSize]                      ; sectors used by FATs
    mov cx, ax

    ; compute location of FAT and store in "ax"
    mov ax, WORD [ReservedSec]              ; adjust for bootsector
          
    ; read FAT into memory (7C00:0200)
    mov bx, 0x0200                          ; copy FAT above bootcode
    call ReadSectors

    ; read image file into memory (0050:0000)
    mov ax, 0x0050
    mov es, ax                              ; destination for image
    mov bx, 0x0000                          ; destination for image
    push bx

;******************************************************************************
;   Load stage2 bootloader
;******************************************************************************

LOAD_IMAGE:
    mov ax, WORD [cluster]                  ; cluster to read
    pop bx                                  ; buffer to read into
    call ClusterLBA                         ; convert cluster to LBA
    xor cx, cx
    mov cl, BYTE [SecPerClus]               ; sectors to read
    call ReadSectors
    push bx
          
    ; compute next cluster
    mov ax, WORD [cluster]                  ; identify current cluster
    mov cx, ax                              ; copy current cluster
    mov dx, ax                              ; copy current cluster
    shr dx, 0x0001                          ; divide by two
    add cx, dx                              ; sum for (3/2)
    mov bx, 0x0200                          ; location of FAT in memory
    add bx, cx                              ; index into FAT
    mov dx, WORD [bx]                       ; read two bytes from FAT
    test ax, 0x0001
    jnz .ODD_CLUSTER
          
.EVEN_CLUSTER:
    and dx, 0000111111111111b               ; take low twelve bits
    jmp .DONE
         
.ODD_CLUSTER:
    shr dx, 0x0004                          ; take high twelve bits
          
.DONE:
    mov WORD [cluster], dx                  ; store new cluster
    cmp dx, 0x0FF0                          ; test for end of file
    jb LOAD_IMAGE
          
DONE:
    mov si, msgCRLF
    call print_string
	mov dl, [bootdevice]
    push WORD 0x0050
    push WORD 0x0000
    retf
          
FAILURE:
    mov si, msgFailure
    call print_string
    mov ah, 0x00
    int 0x16                                ; wait for keypress
    int 0x19                                ; warm boot reset

;******************************************************************************
;   Print String  
;	DS:SI   null-terminated string
;******************************************************************************
print_string:
    mov ah, 0x0E      ; BIOS function 0x0E: teletype
.loop:   
    lodsb             ; grab a byte from SI
    test al, al       ; NUL?
    jz .done          ; if the result is zero: get out
    int 0x10          ; else: print out the character
    jmp .loop
.done:
    ret	

;******************************************************************************
;   Convert CHS to LBA
;   LBA = (cluster - 2) * sectors per cluster
;******************************************************************************
ClusterLBA:
    sub ax, 0x0002                          ; zero base cluster number
    xor cx, cx
    mov cl, BYTE [SecPerClus]               ; convert byte to word
    mul cx
    add ax, WORD [datasector]               ; base data sector
    ret

;******************************************************************************
;   Convert LBA to CHS
;   AX    LBA Address to convert
;
;   absolute sector = (logical sector / sectors per track) + 1
;   absolute head   = (logical sector / sectors per track) MOD number of heads
;   absolute track  = logical sector / (sectors per track * number of heads)
;
;******************************************************************************
Convert_LBA_to_CHS :
    xor dx, dx                              ; prepare dx:ax for operation
    div WORD [SecPerTrack]                  ; calculate
    inc dl                                  ; adjust for sector 0
    mov BYTE [absoluteSector], dl
    xor dx, dx                              ; prepare dx:ax for operation
    div WORD [NumHeads]                     ; calculate
    mov BYTE [absoluteHead], dl
    mov BYTE [absoluteTrack], al
    ret

;******************************************************************************
;   Reads sectors 
;   CX     Number of sectors to read
;   AX     Starting sector
;   ES:BX  Buffer to read to
;******************************************************************************
ReadSectors:
.NEXTSECTOR:
    mov di, 0x0005                          ; five retries for error
.LOOP:
    push ax
    push bx
    push cx
    call Convert_LBA_to_CHS                  ; convert starting sector from LBA to CHS
	mov  ah, 0x02                            ; BIOS read sector
	mov  al, 0x01                            ; read one sector
	mov  ch, BYTE [absoluteTrack]            ; track
	mov  cl, BYTE [absoluteSector]           ; sector
	mov  dh, BYTE [absoluteHead]             ; head
	mov  dl, BYTE [DriveNum]                 ; drive
	int  0x13                                ; invoke BIOS
	jnc  .SUCCESS                            ; test for read error
	xor  ax, ax                              ; BIOS reset disk
	int  0x13                                ; invoke BIOS
	dec  di                                  ; decrement error counter
	pop  cx
	pop  bx
	pop  ax
	jnz  .LOOP                               ; attempt to read again
	int  0x18
.SUCCESS:
    mov  si, msgProgress
    call print_string
    pop  cx
    pop  bx
    pop  ax
    add  bx, WORD [BytesPerSec]              ; queue next buffer
    inc  ax                                  ; queue next sector
    loop .NEXTSECTOR                         ; read next sector
    ret
;******************************************************************************

;	Parameters
absoluteSector db 0x00
absoluteHead   db 0x00
absoluteTrack  db 0x00	
bootdevice  db 0
datasector  dw 0x0000
cluster     dw 0x0000
ImageName   db "BOOT2   SYS"
msgCRLF     db 0x0D, 0x0A, 0x00
msgProgress db "*", 0x00
msgFailure  db 0x0D, 0x0A, "BOOT2.SYS MISSING", 0x0D, 0x0A, 0x00
 
TIMES 510-($-$$) hlt
db 0x55
db 0xAA